import json
import logging
import os
from datetime import datetime, timezone

import boto3

# Configure logging
logger = logging.getLogger()
logger.setLevel(logging.INFO)

# Hardcoded base64 token for basic authentication. Authentication method should be replaced according to your company's security policy.
BASIC_AUTH_TOKEN = os.environ["BASIC_AUTH_TOKEN"]

# The ID of the Cognito's user pool to which the app client will belong to. Required for Client Credentials.
USER_POOL_ID = os.environ["USER_POOL_ID"]

# Should be replaced by your configured Custom Scopes. Required for Client Credentials.
OAUTH_CUSTOM_SCOPES = ["webhook-demo-identifier/json.read"]


def authenticate(headers):
    """
    Performs basic authentication.
    Returns True if the authentication is successful, False otherwise.
    """
    if "authorization" in headers or "Authorization" in headers:
        token = headers.get("authorization") or headers.get("Authorization")
        token = token.split(" ")[1] if token else None
        return token == BASIC_AUTH_TOKEN
    return False


def create_api_key(request_body):
    """Creates a new API Key and associates it with the requested Usage Plans. Return the API Key ID and value."""
    client = boto3.client("apigateway")

    app_name = request_body.get("appName")
    developer = request_body.get("developer")
    apis = request_body.get("apis", [])

    try:
        api_key_response = client.create_api_key(
            name=f"{app_name}-{developer}",
            enabled=True,
            # tags={'string':'string} can be used to distinct API Key created by the webhook
        )

        api_key_id = api_key_response["id"]
        api_key_value = api_key_response["value"]

        # Iterate over APIs in the request body
        for api_data in apis:
            # api_id = api_data.get('id') can be used to verify if API is still related to the Usage Plan
            usage_plans = api_data.get("usagePlans", [])

            # Iterate over usage plans for each API
            for usage_plan in usage_plans:
                usage_plan_id = usage_plan.get("id")

                # Associate API Key with Usage Plan
                client.create_usage_plan_key(
                    usagePlanId=usage_plan_id, keyId=api_key_id, keyType="API_KEY"
                )

        return {
            "statusCode": 200,
            "body": json.dumps(
                {
                    "credentialType": "API_KEY",
                    "apiKeyId": api_key_id,
                    "apiKey": api_key_value,
                }
            ),
        }
    except Exception as e:
        logger.error("create_api_key error: %s", str(e))
        return handle_error(str(e))


def update_with_api_key(request_body):
    """Updates the Usage Plans for a given API Key."""
    client = boto3.client("apigateway")

    apis = request_body.get("apis", [])

    try:
        api_key_id = request_body.get("apiKeyId")

        # Iterate over APIs in the request body
        for api_data in apis:
            # api_id = api_data.get('id') can be used to verify if API is still related to the Usage Plan
            usage_plans = api_data.get("usagePlans", [])

            # Iterate over usage plans for each API
            for usage_plan in usage_plans:
                usage_plan_id = usage_plan.get("id")
                usage_plan_action = usage_plan.get("action")

                if "ADDED" == usage_plan_action:
                    # Associate API Key with Usage Plan
                    client.create_usage_plan_key(
                        usagePlanId=usage_plan_id, keyId=api_key_id, keyType="API_KEY"
                    )
                else:
                    if "REMOVED" == usage_plan_action:
                        client.delete_usage_plan_key(
                            usagePlanId=usage_plan_id, keyId=api_key_id
                        )

        return {
            "statusCode": 204,
        }
    except Exception as e:
        logger.error("update_with_api_key error: %s", str(e))
        return handle_error(str(e))


def update_with_client_credentials(request_body):
    """Creates a Cognito app client and returns the client id and client secret."""
    app_name = request_body.get("appName")
    developer = request_body.get("developer")

    # The name of the app client to create. This is the name you will see in AWS console.
    client_name = f"{app_name}-{developer}"

    cognito_client = boto3.client("cognito-idp")

    try:
        response = cognito_client.create_user_pool_client(
            UserPoolId=USER_POOL_ID,
            ClientName=client_name,
            GenerateSecret=True,
            AllowedOAuthFlowsUserPoolClient=True,
            AllowedOAuthScopes=OAUTH_CUSTOM_SCOPES,
            AllowedOAuthFlows=["client_credentials"],
            SupportedIdentityProviders=["COGNITO"],
        )

        app_client_id = response["UserPoolClient"]["ClientId"]
        app_client_secret = response["UserPoolClient"]["ClientSecret"]

        return {
            "statusCode": 200,
            "body": json.dumps(
                {
                    "credentialType": "CLIENT_CREDENTIALS",
                    "clientId": app_client_id,
                    "clientSecret": app_client_secret,
                }
            ),
        }
    except Exception as e:
        logger.error("create_user_pool_client error: %s", str(e))
        return handle_error(str(e))


def create_cognito_app_client(request_body):
    """Creates a Cognito app client and returns the client id and client secret."""
    app_name = request_body.get("appName")
    developer = request_body.get("developer")

    # The name of the app client to create. This is the name you will see in AWS console.
    client_name = f"{app_name}-{developer}"

    cognito_client = boto3.client("cognito-idp")

    try:
        response = cognito_client.create_user_pool_client(
            UserPoolId=USER_POOL_ID,
            ClientName=client_name,
            GenerateSecret=True,
            AllowedOAuthFlowsUserPoolClient=True,
            AllowedOAuthScopes=OAUTH_CUSTOM_SCOPES,
            AllowedOAuthFlows=["client_credentials"],
            SupportedIdentityProviders=["COGNITO"],
        )

        app_client_id = response["UserPoolClient"]["ClientId"]
        app_client_secret = response["UserPoolClient"]["ClientSecret"]

        return {
            "statusCode": 200,
            "body": json.dumps(
                {
                    "credentialType": "CLIENT_CREDENTIALS",
                    "clientId": app_client_id,
                    "clientSecret": app_client_secret,
                }
            ),
        }
    except Exception as e:
        logger.error("create_user_pool_client error: %s", str(e))
        return handle_error(str(e))


def delete_api_key(request_body):
    """Deletes an API key."""
    client = boto3.client("apigateway")

    api_key_id = request_body.get("apiKeyId")

    try:
        client.delete_api_key(apiKey=api_key_id)

        return {"statusCode": 204}
    except Exception as e:
        logger.error("delete_api_key error: %s", str(e))
        return handle_error(str(e))


def delete_cognito_app_client(request_body):
    """Deletes a Cognito app client."""
    cognito_client = boto3.client("cognito-idp")

    app_client_id = request_body.get("clientId")

    try:
        cognito_client.delete_user_pool_client(
            UserPoolId=USER_POOL_ID, ClientId=app_client_id
        )

        return {"statusCode": 204}
    except Exception as e:
        logger.error("delete_user_pool_client error: %s", str(e))
        return handle_error(str(e))


def check_health():
    """Checks the health of the service."""
    return {"statusCode": 204}


def handler(event, context):
    """
    Handles incoming requests.
    """
    logger.debug("Handling event: %s", event)
    path = event["path"]

    if path != "/v1/health" and not authenticate(event.get("headers", {})):
        return handle_error("Unauthorized", 401)

    endpoint_handlers = {
        ("POST", "/v1/createCredentials"): create_api_key,
        # ("POST", "/v1/createCredentials"): create_cognito_app_client,
        # ("POST", "/v1/updateCredentials"): update_with_api_key,
        ("POST", "/v1/updateCredentials"): update_with_client_credentials,
        ("POST", "/v1/revokeCredentials"): delete_api_key,
        # ("POST", "/v1/revokeCredentials"): delete_cognito_app_client,
        ("GET", "/v1/health"): check_health,
    }

    method = event["httpMethod"]
    handler = endpoint_handlers.get((method, path))

    if handler:
        if path != "/v1/health":
            body = json.loads(event.get("body", "{}"))
            return handler(body)
        else:
            return handler()
    else:
        if (method, path) in endpoint_handlers.keys():
            return handle_error("Method Not Allowed", 405)
        else:
            return handle_error("Not Found", 404)


def current_datetime_utc():
    return datetime.now(timezone.utc).isoformat()


def handle_error(error_message, status_code=500):
    return {
        "statusCode": status_code,
        "body": json.dumps(
            {
                "status": status_code,
                "messages": [error_message],
                "timestamp": current_datetime_utc(),
            }
        ),
    }
